require 'spec_helper'

describe NodeBaseline do
  describe "#carry_over_to_tomorrow" do
    it "should create a new NodeBaseline pointing at the same Node and Baseline, with the next day as its date" do
      report        = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))
      node_baseline = NodeBaseline.create_from_report!(report, report.time)
      new_node_baseline = node_baseline.carry_over_to_tomorrow

      new_node_baseline.baseline_id.should == node_baseline.baseline_id
      new_node_baseline.node_id.should     == node_baseline.node_id
      new_node_baseline.date.should        == node_baseline.date.tomorrow
      new_node_baseline.id.should_not      == node_baseline.id
    end
  end

  describe "#clone_and_apply_changes" do
    it "should copy the Baseline and its associated resources/properties" do
      report        = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))
      node_baseline = NodeBaseline.create_from_report!(report, report.time)
      new_node_baseline = node_baseline.clone_and_apply_changes

      new_node_baseline.node_id.should == node_baseline.node_id
      new_node_baseline.date.should    == node_baseline.date.tomorrow

      new_resources = new_node_baseline.baseline.resources.map do |res|
        res_name = "#{res.resource_type}[#{res.title}]"
        props = res.properties.map {|prop| {prop.name => prop.value}}.inject(&:merge)
      end.inject(&:merge)

      old_resources = node_baseline.baseline.resources.map do |res|
        res_name = "#{res.resource_type}[#{res.title}]"
        props = res.properties.map {|prop| {prop.name => prop.value}}.inject(&:merge)
      end.inject(&:merge)

      new_resources.should == old_resources
    end

    it "should apply any provided changes to the new baseline" do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/baseline.yaml')))
      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))

      node_baseline = NodeBaseline.create_from_report!(baseline_report, report.time)

      Summarizer.build_baseline_summary(report.id, node_baseline)

      differences = node_baseline.differences

      differences.find_by_resource_name('File[/etc/hosts]').accept_change

      changes = differences.find_all_by_accepted(true)

      new_node_baseline = node_baseline.clone_and_apply_changes(changes)

      new_resource = new_node_baseline.baseline.resources.find_by_resource_type_and_title('File', '/etc/hosts')

      new_resource.properties.map {|prop| [prop.name, prop.value]}.should =~ [
        ['ensure',  :file],
        ['content', '{md5}6cd31951036020aee7ff2d27e3953063'],
        ['owner',   '0'],
        ['group',   '0'],
        ['mode',    '644'],
        ['type',    'file'],
        ['ctime',   Time.parse('2011-06-15 09:53:48 -07:00')],
        ['mtime',   Time.parse('2011-06-15 09:53:48 -07:00')]
      ]
    end

    it "should use the latest version of a resource when two changes are accepted" do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/baseline.yaml')))
      report1 = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))
      report2 = Factory(:inspect_report, :host => report1.host, :time => report1.time.tomorrow)

      node_baseline1 = NodeBaseline.create_from_report!(baseline_report, report1.time)
      node_baseline2 = NodeBaseline.create_from_report!(baseline_report, report2.time)

      Summarizer.build_baseline_summary(report1.id, node_baseline1)
      Summarizer.build_baseline_summary(report2.id, node_baseline2)

      node_baseline1.differences.find_by_resource_name('File[/etc/hosts]').accept_change
      node_baseline2.differences.find_by_resource_name('File[/etc/hosts]').accept_change

      changes = BaselineDifference.find_all_by_accepted(true)

      new_node_baseline = node_baseline2.clone_and_apply_changes(changes)

      new_node_baseline.baseline.resources.find_by_resource_type_and_title('File', '/etc/hosts').should_not be_present
    end
  end

  describe "::create_for_next_day" do
    before :each do
      @baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/baseline.yaml')))
      @report1 = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))

      # These need to be *today* for the tests to work.
      @baseline_report.time = (Time.now - 10.minutes)
      @report1.time = (Time.now - 5.minutes)

      @nb1 = NodeBaseline.create_from_report!(@baseline_report, @report1.time)

      Summarizer.build_baseline_summary(@report1.id, @nb1)
    end

    it "should carry over if there are no pending changes" do
      BaselineDifference.find_all_by_accepted(true).count.should == 0

      nb2 = @nb1.create_for_next_day
      nb2.should be_an_instance_of NodeBaseline
      nb2.baseline_id.should == @nb1.baseline_id
      nb2.date.should == @nb1.date.tomorrow
    end

    it "should carry over if there are only rejected pending changes" do
      BaselineDifference.all.each {|bd| bd.reject_change }

      nb2 = @nb1.create_for_next_day
      nb2.should be_an_instance_of NodeBaseline
      nb2.baseline_id.should == @nb1.baseline_id
      nb2.date.should == @nb1.date.tomorrow
    end

    it "should clone and apply if there are only accepted pending changes" do
      BaselineDifference.all.each {|bd| bd.accept_change }

      nb2 = @nb1.create_for_next_day
      nb2.should be_an_instance_of NodeBaseline
      nb2.baseline_id.should_not == @nb1.baseline_id
      nb2.date.should == @nb1.date.tomorrow
    end

    it "should clone and apply if there are a mix of accepted and rejected pending changes" do
      BaselineDifference.count.should >= 2
      bd = BaselineDifference.all
      bd[0].reject_change
      bd[1].accept_change

      nb2 = @nb1.create_for_next_day
      nb2.should be_an_instance_of NodeBaseline
      nb2.baseline_id.should_not == @nb1.baseline_id
      nb2.date.should == @nb1.date.tomorrow
    end
  end

  describe "::create_from_report!" do
    it "should create a NodeBaseline from an inspect report" do
      report = Factory(:inspect_report)

      NodeBaseline.create_from_report!(report, report.time)

      NodeBaseline.count.should == 1
    end

    it "should only allow one entry for a node for a day" do
      report = Factory(:inspect_report)

      NodeBaseline.create_from_report!(report, report.time)
      expect { NodeBaseline.create_from_report!(report, report.time) }.to raise_error(ActiveRecord::RecordInvalid)
    end
  end

  describe ".delete_older_than" do
    it 'should delete NodeBaseline, along with related resources, properties, and differences' do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))
      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)
      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))

      Summarizer.build_baseline_summary(report.id, node_baseline)

      baseline_id = node_baseline.baseline_id

      resource_ids = BaselineResource.find(:all, :select => 'id', :conditions => { :baseline_id => baseline_id }).collect do |resource|
        resource.id
      end
      property_ids = BaselineProperty.find(:all, :select => 'id', :conditions => { :baseline_resource_id => resource_ids }).collect do |property|
        property.id
      end

      BaselineResource.count(:conditions => { :baseline_id => baseline_id }).should > 0
      BaselineProperty.count(:conditions => { :id => property_ids }).should > 0
      BaselineDifference.count(:conditions => { :node_baseline_id => node_baseline.id }).should > 0

      NodeBaseline.delete_older_than(report.time + 1)

      BaselineResource.count(:conditions => { :baseline_id => baseline_id }).should == 0
      BaselineProperty.count(:conditions => { :id => property_ids }).should == 0
      BaselineDifference.count(:conditions => { :node_baseline_id => node_baseline.id }).should == 0
    end

    it 'should gracefully return when no node baselines need to be deleted' do
      NodeBaseline.delete_older_than(Time.utc(1999, 1, 1))
    end

    it 'should delete NodeBaselines, but not any Baselines that are still referenced' do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))
      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)

      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), '../fixtures/reports/inspect.yaml')))

      Summarizer.build_baseline_summary(report.id, node_baseline)

      node_baseline.carry_over_to_tomorrow

      NodeBaseline.count(:conditions => {:baseline_id => node_baseline.baseline_id}).should == 2

      NodeBaseline.delete_older_than(NodeBaseline.maximum(:date))

      NodeBaseline.count(:conditions => {:baseline_id => node_baseline.baseline_id}).should == 1
      Baseline.count(:conditions => {:id => node_baseline.baseline_id}).should == 1
    end
  end

  describe ".baseline_summary_for_group_and_date" do
    it 'should return the counts for all nodes in a group' do
      report_foo = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/foo.yaml")))
      report_bar = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/bar.yaml")))

      baseline_report_foo = Factory(:inspect_report, :host => report_foo.host, :time => report_foo.time + 1.hour)
      baseline_report_bar = Factory(:inspect_report, :host => report_bar.host, :time => report_bar.time + 1.hour)

      node_baseline_foo = NodeBaseline.create_from_report!(baseline_report_foo, baseline_report_foo.time)
      node_baseline_bar = NodeBaseline.create_from_report!(baseline_report_bar, baseline_report_bar.time)

      Summarizer.build_baseline_summary(report_foo.id, node_baseline_foo)
      Summarizer.build_baseline_summary(report_bar.id, node_baseline_bar)

      group = NodeGroup.generate!()
      group.nodes = [report_foo.node, report_bar.node]

      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, true).should == 0
      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, false).should == 0
      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, nil).should == 6

      node_baseline_foo.differences.each {|diff| diff.accept_change}
      node_baseline_bar.differences.each {|diff| diff.reject_change}

      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, true).should == 3
      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, false).should == 3
      NodeBaseline.resource_counts_for_group_and_date(group, node_baseline_foo.date, nil).should == 0
    end

    it 'should return the number of nodes in a group with baselines on the given day' do
      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))

      baseline_report = Factory(:inspect_report, :host => report.host, :time => report.time + 1.hour)

      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)

      Summarizer.build_baseline_summary(report.id, node_baseline)

      group = NodeGroup.generate!()
      group.nodes = [report.node, Node.generate!()]

      NodeBaseline.node_count_for_group_and_date(group, node_baseline.date).should == 1
    end
  end
end
